classdef PulseStreamer_RPCLogger < PulseStreamer.PulseStreamer
    %PULSESTREAMER_RPCLOGGER PulseStreamer class with RPC logging.
    %
    %   This class allows to record all RPC requests and responses into 
    %   a log file.
    %   This class is intended for debugging purpose.
    %   Performance may be reduced when log is enabled and this reduction 
    %   depends on the log size.
    %
    %   Logging is disabled by default and can be enabled by calling the
    %   method "enableLog(log_depth, log_filename)"
    %   Log is operating as a circular buffer with size defined by
    %   "log_depth" parameter. This prevents the file from growing
    %   indefinitely.
    % 
    %   When the logger catch an exception it creates a buffer snapshot
    %   file with the timestamp in the filename. This allows to record
    %   every occurence of transmission errors.
    % 
    %   Log file is a .mat file containing a single cell array with the 
    %   following content {rec_serial, numeric_date, rec_type, rec_data}
    %
    %   The sample code below shows how to use this class. It is a drop-in 
    %   replacement to PulseStreamer class with all functionality plus new 
    %   functions from this class
    %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    %       % Code sample:
    %
    %       % As usually make sure that folder is in the MATLAB path
    %       addpath('PulseStreamer'); 
    %
    %       ipAddress = 'pulsestreamer';
    %    
    %       % Instead of "PulseStreamer"
    %       %      ps = PulseStreamer(ipAddress);
    %       %
    %       % Instantiate "debug.PulseStreamer_RPCLogger"
    %       ps = debug.PulseStreamer_RPCLogger(ipAddress);
    %    
    %       % Enable logging
    %       ps.enableLog(100, 'log_file');
    %    
    %       % From now on all requests, responses and exceptions 
    %       % will be stored in a log file
    %    
    %       % Disable logging
    %       ps.enableLog(0);
    %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    
    properties (SetAccess = private)
        logDepth        % maximum number of requests to be stored
        logFilename     % filename the log will be stored to
        logBuffer       % circular buffer for stored records
        logCounter      % incremental counter of log records
    end
    
    methods
        function obj = PulseStreamer_RPCLogger(ipAddress)
            % Construct an instance of this class
            
            obj = obj@PulseStreamer.PulseStreamer(ipAddress);
            
            obj.printBanner();
            obj.logDepth = 0;
        end
        
        function printBanner(obj)
            fprintf('\n');
            fprintf('************************************************\n');
            fprintf('***     PULSE STREAMER WITH RPC LOGGER       ***\n');
            fprintf('***            Only for debugging!!!         ***\n');
            fprintf('***    Use enableLog(log_depth, filename)    ***\n');
            fprintf('***     to enable log file recording.        ***\n');
            fprintf('************************************************\n');
            fprintf('*   Client version:     %s\n', PulseStreamer.client_version());
            fprintf('*   Firmware version:   %s\n', obj.getFirmwareVersion());
            fprintf('*   Serial number:      %s\n', obj.getSerial());
            fprintf('*   FPGA ID:            %s\n', obj.getFPGAID());
            fprintf('*   Max level duration: %10.0f ns\n', obj.max_step_duration);
            fprintf('*   Max pulse count:    %0.0f\n', obj.max_step_count);
            fprintf('************************************************\n');
        end
        
        function enableLog(obj, log_depth, filename)
            % Enable recording of communication data between client and
            % Pulse Streamer hardware. Both requests and responses are 
            % recorded with maximum number of records equal to "log_depth"
            %
            % Set "log_depth=0" to disable recorder
            
            if log_depth>0
                warning(['Logging PulseStreamer communication to file: ', filename]);
                obj.logDepth = log_depth;
                obj.logFilename = filename;
                obj.logBuffer = cell(log_depth, 4);
            else
                warning('Logging disabled.');
                obj.logDepth = 0;
                obj.logFilename = [];
                obj.logBuffer = [];
            end
            obj.logCounter = 0;
        end
    end
    
    methods (Access = {?PulseStreamer.PulseStreamer})
        function ret = httpRequest(obj, req)
            if obj.logDepth > 0
                obj.log_append('REQ', req); % Log request
            end
            
            try
                ret = httpRequest@PulseStreamer.PulseStreamer(obj, req); % pass to parent
                
            catch ME
                if obj.logDepth > 0
                    % Log exception description
                    obj.log_append('EXC', getReport(ME));
                    obj.save_log_snapshot();
                end
                rethrow(ME);
            end
            
            if obj.logDepth > 0
                obj.log_append('ANS', ret); % Log response
            end
        end
        
        function resp = tcpRequest(obj, req)
            if obj.logDepth > 0
                obj.log_append('REQ', req.serialize()); % Log request
            end
            
            try
                resp = tcpRequest@PulseStreamer.PulseStreamer(obj, req); % pass to parent
                
            catch ME
                if obj.logDepth > 0
                    % Log exception description
                    obj.log_append('EXC', getReport(ME));
                    obj.save_log_snapshot();
                end
                rethrow(ME);
            end
            
            if obj.logDepth > 0
                obj.log_append('ANS', resp.serialize()); % Log response
            end
        end
        
    end
    
    methods (Access = private)
        function log_append(obj, type, data)
            % Add new record to a circular buffer. 
            % If full, the new record replaces the oldest one.
            % Oldest record first, newest record last.
            
            obj.logCounter = obj.logCounter + 1;
            obj.logBuffer = circshift(obj.logBuffer, -1); % shift up
            obj.logBuffer(end,:) = {obj.logCounter, now, type,  data};
            log_buffer = obj.logBuffer; %#ok<NASGU>
            save(obj.logFilename, 'log_buffer');
        end
        
        function save_log_snapshot(obj)
            % Stores current log buffer contents to a file
            % with unique filename. 
            
            [dir_path, fname, ext] = fileparts(obj.logFilename);
            snapshot_file_name = sprintf('%s_%s%s', fname, ...
                datestr(now, 'yyyy-mm-dd_HHMMSS_FFF'), ext);
            log_buffer = obj.logBuffer; %#ok<NASGU>
            save(fullfile(dir_path, snapshot_file_name), 'log_buffer');
        end
    end
    
    methods (Static)
        function convert_log_to_text_file(log_file)
            %CONVERT_LOG_TO_TEXT_FILE Converts log file into a plain text file 
            %
            % Usage:
            %   PulseStreamer.debug.PulseStreamer_RPCLogger.convert_log_to_text_file(filepath)
            
            data = load(log_file, '-mat', 'log_buffer');
            log = data.log_buffer;
            
            [~,fname,~] = fileparts(log_file);
            
            try
                fid = fopen([fname, '.txt'], 'w');
                
                for ii = 1:size(log, 1)
                    log_N = log{ii,1};
                    if ~isempty(log_N) % Skip empty records
                        log_time = datestr(log{ii,2}, 'yyyy-mm-dd HH:MM:SS.FFF');
                        log_action = log{ii,3};
                        log_msg = log{ii,4};
                        fprintf(fid, 'N=%7d "%s" %s: %s\n', ...
                            log_N, log_time, log_action, log_msg);
                    end
                end
                fclose(fid);
            catch ME
                fclose(fid);
                rethrow(ME);
            end
        end
    end
end


