function ps_list = findPulseStreamers(search_serial, recv_callback)
%FINDPULSESTREAMERS: Queries the network for available Pulse Streamers.
% Returns discovered Pulse Streamer IP address and identification strings.

    QUERY_STRING = 'PS-DISCOVERY';
    RESPONSE_PREFIX = 'PS-DISCOVERY-REPLY';
    
    BCAST_IP = '255.255.255.255';
    MCAST_IP = '239.255.255.82';
    PORT = 12345;
    
    if exist('search_serial', 'var')
        match = regexp(search_serial, '((?:[\da-fA-F]{2}[:-]?){5}[\da-fA-F]{2})', 'match', 'once');
        if ~strcmp(match, search_serial)
            throw(MException('PulseStreamer:incorrectSerialInput', [
                'Parameter "search_serial" has incorrect value "%s".' ...
                ' Must have the following format:' ...
                ' "12:34:56:78:9A:BC" or be an empty string.'], search_serial));
        end
    end
    
    bcast_pkt = create_packet(BCAST_IP, PORT, QUERY_STRING);
    mcast_pkt = create_packet(MCAST_IP, PORT, QUERY_STRING);
    
    % Send discovery packed from each available and active IP interface
    sock = {};
    N = 1;
    ifaces = java.net.NetworkInterface.getNetworkInterfaces();
    while ifaces.hasMoreElements()
        iface = ifaces.nextElement();
        if iface.isUp() && ~iface.isLoopback()
            addr_list = iface.getInetAddresses();
            while addr_list.hasMoreElements()
                ipaddr = addr_list.nextElement();
                if isa(ipaddr, 'java.net.Inet4Address')
                    sock{N} = create_datagram_socket(ipaddr); %#ok<AGROW>
                    sock{N}.send(bcast_pkt);
                    sock{N}.send(mcast_pkt);
                    N = N + 1;
                end
            end
        end
    end

    % wait for responses
    pause(0.5);

    % Receive
    ps_list = PulseStreamer.internal.DeviceInfo.empty(0);
    
    for ii = 1:numel(sock)
        try
            while true
                % Read messages from socket
                [addr, msg] = receive_packet(sock{ii});
                
                % Check if this is valid response packet
                if ~strncmp(msg, RESPONSE_PREFIX, numel(RESPONSE_PREFIX))
                    continue
                end
                
                % Check if address is already in the list
                if ~isempty(ps_list) 
                    if ismember(addr, {ps_list(:).ip})
                        continue
                    end
                end
                
                % Strip constant prefix
                msg = replace(msg, RESPONSE_PREFIX, '');
                info = PulseStreamer.internal.DeviceInfo(addr, msg);
                
                % If specific serial was requested
                if exist('search_serial', 'var')
                    % Check if the serial match the requested one.
                    % Ignore this check if search_serial is an emtpy string
                    if ~isempty(search_serial)
                        if ~serial_cmpi(info.serial, search_serial)
                            continue
                        end
                    end
                end
                
                % Call the callback if it was specified
                if exist('recv_callback', 'var')
                    if isa(recv_callback, 'function_handle')
                        try
                            recv_callback(dev_info);
                        catch ME
                            warning(getReport(ME));
                        end
                    end
                end
                
                % Finally add to the list
                ps_list(end+1) = info; %#ok<AGROW>
            end
            
        catch ME
            if ~isprop(ME, 'ExceptionObject')
                rethrow(ME);
            end
            if isa(ME.ExceptionObject, 'java.net.SocketTimeoutException')
                sock{ii}.close();
            else
                rethrow(ME);
            end
        end
    end
end


function ret = serial_cmpi(str1, str2)
    % Return True if serial strings are equivalent.
    str1 = strrep(strrep(str1, ':', ''), '-', '');
    str2 = strrep(strrep(str2, ':', ''), '-', '');
    ret = strcmpi(str1, str2);
end
    
function s = create_datagram_socket(addr)
    % Create datagram socket and set appropriate settings
    s = java.net.DatagramSocket(0, addr);
    s.setSoTimeout(1);
    s.setBroadcast(1);
    s.setReuseAddress(1);
end

function pkt = create_packet(dst_ip, dst_port, msg)
    msg_data = int8(msg);
    msg_len = numel(msg_data);
    ipaddr = java.net.InetAddress.getByName(dst_ip);
    port = uint16(dst_port);
    pkt = java.net.DatagramPacket(msg_data, msg_len, ipaddr, port);
end

function [addr, msg] = receive_packet(sock)
    pkt = java.net.DatagramPacket(zeros(1, 256, 'int8'), 256);
    sock.receive(pkt);
    msg_data = pkt.getData().';
    msg_len = pkt.getLength();
    msg = char(msg_data(1:msg_len));
    addr = char(pkt.getAddress().getHostAddress());
end